#include <stdio.h>
#include <stdlib.h>

struct node {
	int v;
	struct node *next;
};

struct graph {
	int V;
	int E;
	struct node **A;
};

struct graph *creategraph(int nv, int ne)
{
	struct graph *g = malloc(sizeof(struct graph));
	
	if(!g) {
		fprintf(stderr,"Errore di Allocazione\n");
		exit(-1);
	}
	g->E = ne;
	g->V = nv;
	if(!(g->A = calloc(nv,sizeof(struct node *))))  {
		fprintf(stderr,"Errore di Allocazione\n");
		exit(-2);
	}
	return g;
}

struct node *vertexinsert(struct node *p, int k)
{
	struct node *q = malloc(sizeof(struct node));

	if(!q) {
		fprintf(stderr,"Errore di Allocazione\n");
		exit(-3);
	}
	q->v = k;
	q->next = p;
	return q;
}

void readgraph(struct graph *g, FILE *fp)
{
	int i,v1, v2;
	
	for(i = 0; i < g->E; i++) {
		fscanf(fp,"%d %d",&v1,&v2);
		/* inserisco l'arco (v1,v2) solo nella lista corrispondente a v1 */
		g->A[v1-1] = vertexinsert(g->A[v1-1],v2);
	}		
}

/* leggera modifica della funzione dfs1 di graph.c */
void dfs(struct graph *g, int i, char *aux)
{
	struct node *t;

	aux[i] = 1;
	for(t = g->A[i]; t; t = t->next)
		if(!aux[t->v - 1])
			dfs(g,t->v-1,aux);
}

/* restituisce 1 se j e' raggiungibile da i nel grafo g */
/* 0 altrimenti */
/* la funzione e' ottenuta per modifica da dfs di graph.c */
char reachable(struct graph *g, int i, int j)
{
	char *aux = calloc(g->V,sizeof(int));
	char r;

	if(!aux) {
		fprintf(stderr,"Errore di Allocazione\n");
		exit(-4);
	}
	/* visito la componente connessa contenente il vertice i */
	dfs(g,i-1,aux);
	/* se alla fine la posizione corrispondente a j nell'array aux e' settata */
	/* allora j e' stato raggiunto da un cammino originante in i */
	r = aux[j-1];	
	free(aux);
	return r;
}

int main(int ac, char **av)
{
	int i, j, nv, ne;
	struct graph *g;
	FILE *fp;

	if(ac != 2) {
		fprintf(stderr,"Uso: %s <file>\n",av[0]);
		exit(-1);
	};
	if(!(fp = fopen(av[1],"r"))) {
	fprintf(stderr,"File %s non trovato\n",av[1]);
		exit(-2);
	};	

	fscanf(fp,"%d %d", &nv, &ne);
	g = creategraph(nv,ne);
	readgraph(g,fp);

	/* ciclo su tutte le coppie di vertici */
	for(i = 1; i <= nv; i++)
		for(j = 1; j <= nv; j++)
			/* se j e' raggiungibile da i */
			/* stampo la coppia (i,j) */
			if(j != i && reachable(g,i,j))
				printf("(%d,%d) ",i,j);
	return 0;
}
