/* zaino con ripetizione illimitata di oggetti */
#include <stdio.h>
#include <stdlib.h>

#define DIMWORD 20

/* leggo nomi, pesi e valori di n tipi di oggetti dal file fp */
void leggidati(FILE *fp, char *nome, int *peso, int *valore, int n)
{
	int i;

	for(i = 0; i < n; i++)
		/* aritmetica dei puntatori, es: peso + i equivale a &peso[i] */
		/* nome + (DIMWORD * i) equivale a &nome[DIMWORD * i], vale a dire */
		/* equivale all'indirizzo dove memorizzare l'i-esimo nome di tipo di oggetto */
		fscanf(fp,"%s %d %d",nome + (DIMWORD * i), peso + i, valore + i);
}

/* calcola una soluzione ottimale al problema dello zaino */
/* zaino e' un array di 2(limitepeso + 1) interi, dimensionato nel main */
/* i primi (limitepeso + 1) elementi conterranno i valori delle soluzioni dei sottoproblemi */
/* gli ultimi (limitepeso + 1) elementi permetterranno di ricostruire la struttura della soluzione. */
/* questi ultimi elementi sono indirizzati come elementi dell'array "quale" di (limitepeso + 1) elementi */
int *calcolazaino(int *peso, int *valore, int n, int limitepeso)
{
	
	int *zaino,*quale;
	int i,j,k,v;

	/* alloco in memoria dinamica l'array zaino */
	if(!(zaino = calloc(2*(limitepeso + 1),sizeof(int)))) {
		fprintf(stderr, "Errore nell'allocazione della memoria\n");
		exit(-3);
	}
	/* assegno a quale l'indirizzo del (limitepeso + 1)esimo elemento di zaino */
	/* in questo modo quale si comporta come il nome di un array di (limitepeso + 1) elementi */
	/* Aritmetica degli indirizzi: zaino + limitepeso + 1 equivale a &zaino[limitepeso + 1] */
	quale = zaino + limitepeso + 1;
	/* ciclo sui limiti di peso (le capacita' degli zaini) da 0 a limitepeso (incluso)*/
	for(i = 0; i <= limitepeso; i++)
	{
		/* ciclo sui tipi di oggetti */
		for(j = 0; j < n; j++) {
			/* calcolo la capacita' di uno zaino ottenuto sommando */
			/* alla capacita' i il peso del tipo di oggetto j */
			k = i + peso[j];
			/* calcolo il valore cha avrebbe lo zaino di capacita' k */
			/* se fosse ottenuto da zaino[i] aggiungendo un oggetto di tipo j */
			v = zaino[i] + valore[j];
			/* se k non eccede la capacita' massima e */
			/* se la soluzione ottenuta aggiungendo a zaino[i] un oggetto di tipo j */
			/* e' migliore della soluzione attuale per zaino[k] */
			/* allora aggiorno la soluzione */
			if(k <= limitepeso && zaino[k] < v) {
				zaino[k] = v;
				/* memorizzo che l'ultimo oggetto inserito per questa soluzione di zaino[k] */
				/* e' un oggetto di tipo j */
				quale[k] = j;
			}	
		}
	}
	/* restituisco lo zaino costruito alla fine del ciclo */
	return zaino;
}

/* stampa la soluzione ottima contenuta contenuta nell'array zaino */
void stampazaino(int *zaino, int *peso, char *nome, int n, int limitepeso)
{
	int i;
	/* quale punta alla seconda parte dell'array zaino, come nella funzione calcolazaino */
	int *quale = zaino + limitepeso + 1;
	/* array ausiliario per memorizzare quanti oggetti di ogni tipo sono contenuti nello zaino */
	int *quanto = calloc(n,sizeof(int));

	if(!quanto) {
		fprintf(stderr, "Errore nell'allocazione della memoria\n");
		exit(-3);
	}

	/* il valore totale e' contenuto in zaino[limitepeso] */
	printf("Il valore totale dello zaino e' %d\n",zaino[limitepeso]);
	printf("Lo zaino contiene: \n");

	/* ciclo su i da limitepeso, sottraggo a i il peso dell'ultimo oggetto aggiunto per ottenere zaino[i] */
	for(i = limitepeso; i; i -= peso[quale[i]])
		/* l'ultimo oggetto di zaino[i] e' di tipo quale[i], dunque incremento il */
		/* valore del contatore quanto[quale[i]] */
		quanto[quale[i]]++;

	/* ciclo di stampa finale degli oggetti contenuti */
	for(i = 0; i < n; i++)
		if(quanto[i])
			/* aritmetica degli indirizzi: nome + i * DIMWORD equivale a &nome[i * DIMWORD], */
			/* vale a dire, l'i-esimo nome di tipo di oggetto */
			printf("%d di %s \n",quanto[i],nome + i * DIMWORD);
	putchar('\n');
	/* libero la memoria utilizzata */
	free(quanto);
}

int main(int ac, char **av)
{
	FILE *fp;
	int n, limitepeso;
	int *peso, *valore, *zaino;
	char *nome;
	

	if(ac != 2) {
		fprintf(stderr, "Uso: %s <file>\n", av[0]);
		exit(-1);
	}
	if(!(fp = fopen(av[1],"r"))) {
		fprintf(stderr, "Errore: non riesco ad aprire il file %s\n", av[1]);
		exit(-2);
	}

	fscanf(fp,"%d%d",&limitepeso,&n);
	peso = calloc(n,sizeof(int));
	valore = calloc(n,sizeof(int));
	nome = calloc(n, DIMWORD + 1);
	
	if(!peso || !valore || !nome) {
		fprintf(stderr, "Errore nell'allocazione della memoria\n");
		exit(-3);
	}

	leggidati(fp,nome,peso,valore,n);
	zaino = calcolazaino(peso,valore,n,limitepeso);
	stampazaino(zaino,peso,nome,n,limitepeso);

	return 0;
}
